"use strict";
// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.
//
// Original code forked from https://github.com/Quramy/ts-graphql-plugin
Object.defineProperty(exports, "__esModule", { value: true });
var embeddedSupport_1 = require("./embeddedSupport");
var vscode_html_languageservice_1 = require("vscode-html-languageservice");
var vscode_css_languageservice_1 = require("vscode-css-languageservice");
var vscode = require("vscode-languageserver-types");
var config_1 = require("./config");
var emptyCompletionList = {
    isIncomplete: false,
    items: [],
};
var cssErrorCode = 9999;
var CompletionsCache = /** @class */ (function () {
    function CompletionsCache() {
    }
    CompletionsCache.prototype.getCached = function (context, position) {
        if (this._completions
            && context.fileName === this._cachedCompletionsFile
            && this._cachedCompletionsPosition && arePositionsEqual(position, this._cachedCompletionsPosition)
            && context.text === this._cachedCompletionsContent) {
            return this._completions;
        }
        return undefined;
    };
    CompletionsCache.prototype.updateCached = function (context, position, completions) {
        this._cachedCompletionsFile = context.fileName;
        this._cachedCompletionsPosition = position;
        this._cachedCompletionsContent = context.text;
        this._completions = completions;
    };
    return CompletionsCache;
}());
var HtmlTemplateLanguageService = /** @class */ (function () {
    function HtmlTemplateLanguageService(typescript, configuration, logger // tslint:disable-line
    ) {
        this.typescript = typescript;
        this.configuration = configuration;
        this.logger = logger;
        this._completionsCache = new CompletionsCache();
    }
    Object.defineProperty(HtmlTemplateLanguageService.prototype, "htmlLanguageService", {
        get: function () {
            if (!this._htmlLanguageService) {
                this._htmlLanguageService = vscode_html_languageservice_1.getLanguageService();
            }
            return this._htmlLanguageService;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(HtmlTemplateLanguageService.prototype, "cssLanguageService", {
        get: function () {
            if (!this._cssLanguageService) {
                this._cssLanguageService = vscode_css_languageservice_1.getCSSLanguageService();
            }
            return this._cssLanguageService;
        },
        enumerable: true,
        configurable: true
    });
    HtmlTemplateLanguageService.prototype.getCompletionsAtPosition = function (context, position) {
        var items = this.getCompletionItems(context, position);
        return translateCompletionItemsToCompletionInfo(this.typescript, items);
    };
    HtmlTemplateLanguageService.prototype.getCompletionEntryDetails = function (context, position, name) {
        var items = this.getCompletionItems(context, position).items;
        var item = items.find(function (x) { return x.label === name; });
        if (!item) {
            return {
                name: name,
                kind: this.typescript.ScriptElementKind.unknown,
                kindModifiers: '',
                tags: [],
                displayParts: toDisplayParts(name),
                documentation: [],
            };
        }
        return translateCompletionItemsToCompletionEntryDetails(this.typescript, item);
    };
    HtmlTemplateLanguageService.prototype.getQuickInfoAtPosition = function (context, position) {
        var htmlDoc = this.createVirtualDocument(context);
        var documentRegions = embeddedSupport_1.getDocumentRegions(this.htmlLanguageService, htmlDoc);
        var languageId = documentRegions.getLanguageAtPosition(position);
        var hover;
        switch (languageId) {
            case 'html':
                var html = this.htmlLanguageService.parseHTMLDocument(htmlDoc);
                hover = this.htmlLanguageService.doHover(htmlDoc, position, html);
                break;
            case 'css':
                var cssDoc = this.createCssVirtualDocument(context);
                var stylesheet = this.cssLanguageService.parseStylesheet(cssDoc);
                hover = this.cssLanguageService.doHover(cssDoc, position, stylesheet);
                break;
            default:
                hover = undefined;
                break;
        }
        if (hover) {
            return this.translateHover(hover, position, context, languageId);
        }
        return undefined;
    };
    HtmlTemplateLanguageService.prototype.getFormattingEditsForRange = function (context, start, end, settings) {
        var _this = this;
        if (!this.configuration.format.enabled) {
            return [];
        }
        var doc = this.createVirtualDocument(context);
        // Make sure we don't get rid of leading newline
        var leading = context.text.match(/^\s*\n/);
        if (leading) {
            start += leading[0].length;
        }
        // or any trailing newlines
        var trailing = context.text.match(/\n\s*$/);
        if (trailing) {
            end -= trailing[0].length;
        }
        if (end <= start) {
            return [];
        }
        var range = this.toVsRange(context, start, end);
        var edits = this.htmlLanguageService.format(doc, range, {
            tabSize: settings.tabSize,
            insertSpaces: !!settings.convertTabsToSpaces,
            wrapLineLength: 120,
            unformatted: '',
            contentUnformatted: 'pre,code,textarea',
            indentInnerHtml: false,
            preserveNewLines: true,
            maxPreserveNewLines: null,
            indentHandlebars: false,
            endWithNewline: false,
            extraLiners: 'head, body, /html',
            wrapAttributes: 'auto',
        });
        return edits.map(function (vsedit) {
            return {
                span: _this.toTsSpan(context, vsedit.range),
                newText: vsedit.newText,
            };
        });
    };
    HtmlTemplateLanguageService.prototype.getSignatureHelpItemsAtPosition = function (_context, _position) {
        // Html does not support sig help
        return undefined;
    };
    HtmlTemplateLanguageService.prototype.getOutliningSpans = function (context) {
        var _this = this;
        var doc = this.createVirtualDocument(context);
        var ranges = this.htmlLanguageService.getFoldingRanges(doc);
        return ranges.map(function (range) { return _this.translateOutliningSpan(context, range); });
    };
    HtmlTemplateLanguageService.prototype.getSemanticDiagnostics = function (context) {
        var doc = this.createVirtualDocument(context);
        var documentRegions = embeddedSupport_1.getDocumentRegions(this.htmlLanguageService, doc);
        var embeddedDoc = documentRegions.getEmbeddedDocument('css');
        var stylesheet = this.cssLanguageService.parseStylesheet(embeddedDoc);
        return this.translateDiagnostics(this.cssLanguageService.doValidation(embeddedDoc, stylesheet), doc, context, context.text).filter(function (x) { return !!x; });
    };
    HtmlTemplateLanguageService.prototype.getSupportedCodeFixes = function () {
        return [cssErrorCode];
    };
    HtmlTemplateLanguageService.prototype.getCodeFixesAtPosition = function (context, start, end, errorCodes, format) {
        var range = this.toVsRange(context, start, end);
        var doc = this.createVirtualDocument(context);
        var documentRegions = embeddedSupport_1.getDocumentRegions(this.htmlLanguageService, doc);
        var embeddedDoc = documentRegions.getEmbeddedDocument('css');
        var stylesheet = this.cssLanguageService.parseStylesheet(embeddedDoc);
        var diagnostics = this.cssLanguageService.doValidation(embeddedDoc, stylesheet)
            .filter(function (diagnostic) { return overlaps(diagnostic.range, range); });
        return this.translateCodeActions(context, this.cssLanguageService.doCodeActions(doc, range, { diagnostics: diagnostics }, stylesheet));
    };
    HtmlTemplateLanguageService.prototype.toVsRange = function (context, start, end) {
        return {
            start: context.toPosition(start),
            end: context.toPosition(end),
        };
    };
    HtmlTemplateLanguageService.prototype.toTsSpan = function (context, range) {
        var editStart = context.toOffset(range.start);
        var editEnd = context.toOffset(range.end);
        return {
            start: editStart,
            length: editEnd - editStart,
        };
    };
    HtmlTemplateLanguageService.prototype.createVirtualDocument = function (context) {
        var contents = context.text;
        return {
            uri: 'untitled://embedded.html',
            languageId: 'html',
            version: 1,
            getText: function () { return contents; },
            positionAt: function (offset) {
                return context.toPosition(offset);
            },
            offsetAt: function (p) {
                return context.toOffset(p);
            },
            lineCount: contents.split(/n/g).length + 1,
        };
    };
    HtmlTemplateLanguageService.prototype.createCssVirtualDocument = function (context) {
        var contents = context.text;
        return {
            uri: 'untitled://embedded.css',
            languageId: 'css',
            version: 1,
            getText: function () { return contents; },
            positionAt: function (offset) {
                return context.toPosition(offset);
            },
            offsetAt: function (p) {
                return context.toOffset(p);
            },
            lineCount: contents.split(/n/g).length + 1,
        };
    };
    HtmlTemplateLanguageService.prototype.getCompletionItems = function (context, position) {
        var cached = this._completionsCache.getCached(context, position);
        if (cached) {
            return cached;
        }
        var htmlDoc = this.createVirtualDocument(context);
        var documentRegions = embeddedSupport_1.getDocumentRegions(this.htmlLanguageService, htmlDoc);
        var languageId = documentRegions.getLanguageAtPosition(position);
        var completions;
        switch (languageId) {
            case 'html':
                var html = this.htmlLanguageService.parseHTMLDocument(htmlDoc);
                completions = this.htmlLanguageService.doComplete(htmlDoc, position, html) || emptyCompletionList;
                break;
            case 'css':
                var cssDoc = this.createCssVirtualDocument(context);
                var stylesheet = this.cssLanguageService.parseStylesheet(cssDoc);
                completions = this.cssLanguageService.doComplete(cssDoc, position, stylesheet) || emptyCompletionList;
                break;
            default:
                completions = emptyCompletionList;
                break;
        }
        this._completionsCache.updateCached(context, position, completions);
        return completions;
    };
    HtmlTemplateLanguageService.prototype.translateHover = function (hover, position, context, languageId) {
        var header = [];
        var docs = [];
        var convertPart = function (hoverContents) {
            if (typeof hoverContents === 'string') {
                docs.push({ kind: 'unknown', text: hoverContents });
            }
            else if (Array.isArray(hoverContents)) {
                hoverContents.forEach(convertPart);
            }
            else {
                if (languageId === 'css') {
                    // When inside a <style> block and hovering over the first
                    // property (i.e. <style> a.myclass { ... } ) the value
                    // contains the <style> tag which looks very odd and is not
                    // consistent with hovers in html/css files. We therefore
                    // remove it and any additional spaces to give it a
                    // consistent look and feel.
                    var removeStyleDeclaration = hoverContents.value.replace(/(<style>\n)\s+/gi, '');
                    header.push({ kind: 'unknown', text: removeStyleDeclaration });
                }
                else {
                    header.push({ kind: 'unknown', text: hoverContents.value });
                }
            }
        };
        convertPart(hover.contents);
        var start = context.toOffset(hover.range ? hover.range.start : position);
        return {
            kind: this.typescript.ScriptElementKind.string,
            kindModifiers: '',
            textSpan: {
                start: start,
                length: hover.range ? context.toOffset(hover.range.end) - start : 1,
            },
            displayParts: header,
            documentation: docs,
            tags: [],
        };
    };
    HtmlTemplateLanguageService.prototype.translateOutliningSpan = function (context, range) {
        var startOffset = context.toOffset({ line: range.startLine, character: range.startCharacter || 0 });
        var endOffset = context.toOffset({ line: range.endLine, character: range.endCharacter || 0 });
        var span = {
            start: startOffset,
            length: endOffset - startOffset,
        };
        return {
            autoCollapse: false,
            kind: this.typescript.OutliningSpanKind.Code,
            bannerText: '',
            textSpan: span,
            hintSpan: span,
        };
    };
    HtmlTemplateLanguageService.prototype.translateTextEditToFileTextChange = function (context, textEdit) {
        var start = context.toOffset(textEdit.range.start);
        var end = context.toOffset(textEdit.range.end);
        return {
            fileName: context.fileName,
            textChanges: [{
                    newText: textEdit.newText,
                    span: {
                        start: start,
                        length: end - start,
                    },
                }],
        };
    };
    HtmlTemplateLanguageService.prototype.translateCodeActions = function (context, codeActions) {
        var _this = this;
        var actions = [];
        for (var _i = 0, codeActions_1 = codeActions; _i < codeActions_1.length; _i++) {
            var vsAction = codeActions_1[_i];
            if (vsAction.command !== '_css.applyCodeAction') {
                continue;
            }
            var edits = vsAction.arguments && vsAction.arguments[2];
            if (edits) {
                actions.push({
                    description: vsAction.title,
                    changes: edits.map(function (edit) { return _this.translateTextEditToFileTextChange(context, edit); }),
                });
            }
        }
        return actions;
    };
    HtmlTemplateLanguageService.prototype.translateDiagnostics = function (diagnostics, doc, context, content) {
        var _this = this;
        var sourceFile = context.node.getSourceFile();
        return diagnostics.map(function (diag) {
            return _this.translateDiagnostic(diag, sourceFile, doc, context, content);
        });
    };
    HtmlTemplateLanguageService.prototype.translateDiagnostic = function (diagnostic, file, doc, context, content) {
        // Make sure returned error is within the real document
        if (diagnostic.range.start.line === 0
            || diagnostic.range.start.line > doc.lineCount
            || diagnostic.range.start.character >= content.length) {
            return undefined;
        }
        var start = context.toOffset(diagnostic.range.start);
        var length = context.toOffset(diagnostic.range.end) - start;
        var code = typeof diagnostic.code === 'number' ? diagnostic.code : cssErrorCode;
        return {
            code: code,
            messageText: diagnostic.message,
            category: translateSeverity(this.typescript, diagnostic.severity),
            file: file,
            start: start,
            length: length,
            source: config_1.pluginName,
        };
    };
    return HtmlTemplateLanguageService;
}());
exports.default = HtmlTemplateLanguageService;
function translateCompletionItemsToCompletionInfo(typescript, items) {
    return {
        isGlobalCompletion: false,
        isMemberCompletion: false,
        isNewIdentifierLocation: false,
        entries: items.items.map(function (x) { return translateCompetionEntry(typescript, x); }),
    };
}
function translateCompletionItemsToCompletionEntryDetails(typescript, item) {
    return {
        name: item.label,
        kindModifiers: 'declare',
        kind: item.kind ? translateionCompletionItemKind(typescript, item.kind) : typescript.ScriptElementKind.unknown,
        displayParts: toDisplayParts(item.detail),
        documentation: toDisplayParts(item.documentation),
        tags: [],
    };
}
function translateCompetionEntry(typescript, item) {
    return {
        name: item.label,
        kindModifiers: '',
        kind: item.kind ? translateionCompletionItemKind(typescript, item.kind) : typescript.ScriptElementKind.unknown,
        sortText: '0',
    };
}
function translateionCompletionItemKind(typescript, kind) {
    switch (kind) {
        case vscode.CompletionItemKind.Method:
            return typescript.ScriptElementKind.memberFunctionElement;
        case vscode.CompletionItemKind.Function:
            return typescript.ScriptElementKind.functionElement;
        case vscode.CompletionItemKind.Constructor:
            return typescript.ScriptElementKind.constructorImplementationElement;
        case vscode.CompletionItemKind.Field:
        case vscode.CompletionItemKind.Variable:
            return typescript.ScriptElementKind.variableElement;
        case vscode.CompletionItemKind.Class:
            return typescript.ScriptElementKind.classElement;
        case vscode.CompletionItemKind.Interface:
            return typescript.ScriptElementKind.interfaceElement;
        case vscode.CompletionItemKind.Module:
            return typescript.ScriptElementKind.moduleElement;
        case vscode.CompletionItemKind.Property:
            return typescript.ScriptElementKind.memberVariableElement;
        case vscode.CompletionItemKind.Unit:
        case vscode.CompletionItemKind.Value:
            return typescript.ScriptElementKind.constElement;
        case vscode.CompletionItemKind.Enum:
            return typescript.ScriptElementKind.enumElement;
        case vscode.CompletionItemKind.Keyword:
            return typescript.ScriptElementKind.keyword;
        case vscode.CompletionItemKind.Color:
            return typescript.ScriptElementKind.constElement;
        case vscode.CompletionItemKind.Reference:
            return typescript.ScriptElementKind.alias;
        case vscode.CompletionItemKind.File:
            return typescript.ScriptElementKind.moduleElement;
        case vscode.CompletionItemKind.Snippet:
        case vscode.CompletionItemKind.Text:
        default:
            return typescript.ScriptElementKind.unknown;
    }
}
function translateSeverity(typescript, severity) {
    switch (severity) {
        case vscode.DiagnosticSeverity.Information:
        case vscode.DiagnosticSeverity.Hint:
            return typescript.DiagnosticCategory.Message;
        case vscode.DiagnosticSeverity.Warning:
            return typescript.DiagnosticCategory.Warning;
        case vscode.DiagnosticSeverity.Error:
        default:
            return typescript.DiagnosticCategory.Error;
    }
}
function toDisplayParts(text) {
    if (!text) {
        return [];
    }
    return [{
            kind: 'text',
            text: typeof text === 'string' ? text : text.value,
        }];
}
function arePositionsEqual(left, right) {
    return left.line === right.line && left.character === right.character;
}
function isAfter(left, right) {
    return right.line > left.line || (right.line === left.line && right.character >= left.character);
}
function overlaps(a, b) {
    return !isAfter(a.end, b.start) && !isAfter(b.end, a.start);
}
//# sourceMappingURL=html-template-language-service.js.map